"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sarus_data_spec.protobuf.path_pb2
import sarus_data_spec.protobuf.scalar_pb2
import sarus_data_spec.protobuf.type_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Transform(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    class Spec(google.protobuf.message.Message):
        """Definitions"""
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        IDENTITY_FIELD_NUMBER: builtins.int
        VARIABLE_FIELD_NUMBER: builtins.int
        COMPOSED_FIELD_NUMBER: builtins.int
        PROJECT_FIELD_NUMBER: builtins.int
        FILTER_FIELD_NUMBER: builtins.int
        SHUFFLE_FIELD_NUMBER: builtins.int
        JOIN_FIELD_NUMBER: builtins.int
        CAST_FIELD_NUMBER: builtins.int
        SAMPLE_FIELD_NUMBER: builtins.int
        USER_SETTINGS_FIELD_NUMBER: builtins.int
        PROTECT_DATASET_FIELD_NUMBER: builtins.int
        EXTERNAL_FIELD_NUMBER: builtins.int
        SYNTHETIC_FIELD_NUMBER: builtins.int
        TRANSCODE_FIELD_NUMBER: builtins.int
        INVERSE_TRANSCODE_FIELD_NUMBER: builtins.int
        GET_ITEM_FIELD_NUMBER: builtins.int
        PROTECTED_PATHS_FIELD_NUMBER: builtins.int
        AUTOMATIC_USER_SETTINGS_FIELD_NUMBER: builtins.int
        PUBLIC_PATHS_FIELD_NUMBER: builtins.int
        ASSIGN_BUDGET_FIELD_NUMBER: builtins.int
        AUTOMATIC_BUDGET_FIELD_NUMBER: builtins.int
        ATTRIBUTE_BUDGET_FIELD_NUMBER: builtins.int
        SD_BUDGET_FIELD_NUMBER: builtins.int
        DERIVE_SEED_FIELD_NUMBER: builtins.int
        GROUP_BY_PE_FIELD_NUMBER: builtins.int
        SAMPLING_RATIOS_FIELD_NUMBER: builtins.int
        SELECT_SQL_FIELD_NUMBER: builtins.int
        EXTRACT_FIELD_NUMBER: builtins.int
        RELATIONSHIP_SPEC_FIELD_NUMBER: builtins.int
        DIFFERENTIATED_SAMPLE_FIELD_NUMBER: builtins.int
        @property
        def identity(self) -> global___Transform.Identity: ...
        @property
        def variable(self) -> global___Transform.Variable: ...
        @property
        def composed(self) -> global___Transform.Composed: ...
        @property
        def project(self) -> global___Transform.Project: ...
        @property
        def filter(self) -> global___Transform.Filter: ...
        @property
        def shuffle(self) -> global___Transform.Shuffle: ...
        @property
        def join(self) -> global___Transform.Join: ...
        @property
        def cast(self) -> global___Transform.Cast: ...
        @property
        def sample(self) -> global___Transform.Sample: ...
        @property
        def user_settings(self) -> global___Transform.UserSettings: ...
        @property
        def protect_dataset(self) -> global___Transform.Protect: ...
        @property
        def external(self) -> global___Transform.External:
            """np transforms, pd transforms,..."""
            pass
        @property
        def synthetic(self) -> global___Transform.Synthetic: ...
        @property
        def transcode(self) -> global___Transform.Transcode: ...
        @property
        def inverse_transcode(self) -> global___Transform.InverseTranscode: ...
        @property
        def get_item(self) -> global___Transform.GetItem: ...
        @property
        def protected_paths(self) -> global___Transform.ProtectedPaths: ...
        @property
        def automatic_user_settings(self) -> global___Transform.AutomaticUserSettings: ...
        @property
        def public_paths(self) -> global___Transform.PublicPaths: ...
        @property
        def assign_budget(self) -> global___Transform.AssignBudget: ...
        @property
        def automatic_budget(self) -> global___Transform.AutomaticBudget: ...
        @property
        def attribute_budget(self) -> global___Transform.AttributesBudget: ...
        @property
        def sd_budget(self) -> global___Transform.SDBudget: ...
        @property
        def derive_seed(self) -> global___Transform.DeriveSeed: ...
        @property
        def group_by_pe(self) -> global___Transform.GroupByPE: ...
        @property
        def sampling_ratios(self) -> global___Transform.SamplingRatios: ...
        @property
        def select_sql(self) -> global___Transform.SelectSql: ...
        @property
        def extract(self) -> global___Transform.Extract: ...
        @property
        def relationship_spec(self) -> global___Transform.RelationshipSpec: ...
        @property
        def differentiated_sample(self) -> global___Transform.DifferentiatedSample: ...
        def __init__(self,
            *,
            identity : typing.Optional[global___Transform.Identity] = ...,
            variable : typing.Optional[global___Transform.Variable] = ...,
            composed : typing.Optional[global___Transform.Composed] = ...,
            project : typing.Optional[global___Transform.Project] = ...,
            filter : typing.Optional[global___Transform.Filter] = ...,
            shuffle : typing.Optional[global___Transform.Shuffle] = ...,
            join : typing.Optional[global___Transform.Join] = ...,
            cast : typing.Optional[global___Transform.Cast] = ...,
            sample : typing.Optional[global___Transform.Sample] = ...,
            user_settings : typing.Optional[global___Transform.UserSettings] = ...,
            protect_dataset : typing.Optional[global___Transform.Protect] = ...,
            external : typing.Optional[global___Transform.External] = ...,
            synthetic : typing.Optional[global___Transform.Synthetic] = ...,
            transcode : typing.Optional[global___Transform.Transcode] = ...,
            inverse_transcode : typing.Optional[global___Transform.InverseTranscode] = ...,
            get_item : typing.Optional[global___Transform.GetItem] = ...,
            protected_paths : typing.Optional[global___Transform.ProtectedPaths] = ...,
            automatic_user_settings : typing.Optional[global___Transform.AutomaticUserSettings] = ...,
            public_paths : typing.Optional[global___Transform.PublicPaths] = ...,
            assign_budget : typing.Optional[global___Transform.AssignBudget] = ...,
            automatic_budget : typing.Optional[global___Transform.AutomaticBudget] = ...,
            attribute_budget : typing.Optional[global___Transform.AttributesBudget] = ...,
            sd_budget : typing.Optional[global___Transform.SDBudget] = ...,
            derive_seed : typing.Optional[global___Transform.DeriveSeed] = ...,
            group_by_pe : typing.Optional[global___Transform.GroupByPE] = ...,
            sampling_ratios : typing.Optional[global___Transform.SamplingRatios] = ...,
            select_sql : typing.Optional[global___Transform.SelectSql] = ...,
            extract : typing.Optional[global___Transform.Extract] = ...,
            relationship_spec : typing.Optional[global___Transform.RelationshipSpec] = ...,
            differentiated_sample : typing.Optional[global___Transform.DifferentiatedSample] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"assign_budget",b"assign_budget",u"attribute_budget",b"attribute_budget",u"automatic_budget",b"automatic_budget",u"automatic_user_settings",b"automatic_user_settings",u"cast",b"cast",u"composed",b"composed",u"derive_seed",b"derive_seed",u"differentiated_sample",b"differentiated_sample",u"external",b"external",u"extract",b"extract",u"filter",b"filter",u"get_item",b"get_item",u"group_by_pe",b"group_by_pe",u"identity",b"identity",u"inverse_transcode",b"inverse_transcode",u"join",b"join",u"project",b"project",u"protect_dataset",b"protect_dataset",u"protected_paths",b"protected_paths",u"public_paths",b"public_paths",u"relationship_spec",b"relationship_spec",u"sample",b"sample",u"sampling_ratios",b"sampling_ratios",u"sd_budget",b"sd_budget",u"select_sql",b"select_sql",u"shuffle",b"shuffle",u"spec",b"spec",u"synthetic",b"synthetic",u"transcode",b"transcode",u"user_settings",b"user_settings",u"variable",b"variable"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"assign_budget",b"assign_budget",u"attribute_budget",b"attribute_budget",u"automatic_budget",b"automatic_budget",u"automatic_user_settings",b"automatic_user_settings",u"cast",b"cast",u"composed",b"composed",u"derive_seed",b"derive_seed",u"differentiated_sample",b"differentiated_sample",u"external",b"external",u"extract",b"extract",u"filter",b"filter",u"get_item",b"get_item",u"group_by_pe",b"group_by_pe",u"identity",b"identity",u"inverse_transcode",b"inverse_transcode",u"join",b"join",u"project",b"project",u"protect_dataset",b"protect_dataset",u"protected_paths",b"protected_paths",u"public_paths",b"public_paths",u"relationship_spec",b"relationship_spec",u"sample",b"sample",u"sampling_ratios",b"sampling_ratios",u"sd_budget",b"sd_budget",u"select_sql",b"select_sql",u"shuffle",b"shuffle",u"spec",b"spec",u"synthetic",b"synthetic",u"transcode",b"transcode",u"user_settings",b"user_settings",u"variable",b"variable"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"spec",b"spec"]) -> typing.Optional[typing_extensions.Literal["identity","variable","composed","project","filter","shuffle","join","cast","sample","user_settings","protect_dataset","external","synthetic","transcode","inverse_transcode","get_item","protected_paths","automatic_user_settings","public_paths","assign_budget","automatic_budget","attribute_budget","sd_budget","derive_seed","group_by_pe","sampling_ratios","select_sql","extract","relationship_spec","differentiated_sample"]]: ...

    class External(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class OpIdentifier(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            STD_FIELD_NUMBER: builtins.int
            PANDAS_FIELD_NUMBER: builtins.int
            NUMPY_FIELD_NUMBER: builtins.int
            TENSORFLOW_FIELD_NUMBER: builtins.int
            SKLEARN_FIELD_NUMBER: builtins.int
            PANDAS_PROFILING_FIELD_NUMBER: builtins.int
            XGBOOST_FIELD_NUMBER: builtins.int
            SKOPT_FIELD_NUMBER: builtins.int
            IMBLEARN_FIELD_NUMBER: builtins.int
            @property
            def std(self) -> global___Transform.External.Std: ...
            @property
            def pandas(self) -> global___Transform.External.Pandas: ...
            @property
            def numpy(self) -> global___Transform.External.Numpy: ...
            @property
            def tensorflow(self) -> global___Transform.External.Tensorflow: ...
            @property
            def sklearn(self) -> global___Transform.External.Sklearn: ...
            @property
            def pandas_profiling(self) -> global___Transform.External.PandasProfiling: ...
            @property
            def xgboost(self) -> global___Transform.External.XGBoost: ...
            @property
            def skopt(self) -> global___Transform.External.Skopt: ...
            @property
            def imblearn(self) -> global___Transform.External.Imblearn: ...
            def __init__(self,
                *,
                std : typing.Optional[global___Transform.External.Std] = ...,
                pandas : typing.Optional[global___Transform.External.Pandas] = ...,
                numpy : typing.Optional[global___Transform.External.Numpy] = ...,
                tensorflow : typing.Optional[global___Transform.External.Tensorflow] = ...,
                sklearn : typing.Optional[global___Transform.External.Sklearn] = ...,
                pandas_profiling : typing.Optional[global___Transform.External.PandasProfiling] = ...,
                xgboost : typing.Optional[global___Transform.External.XGBoost] = ...,
                skopt : typing.Optional[global___Transform.External.Skopt] = ...,
                imblearn : typing.Optional[global___Transform.External.Imblearn] = ...,
                ) -> None: ...
            def HasField(self, field_name: typing_extensions.Literal[u"imblearn",b"imblearn",u"numpy",b"numpy",u"op",b"op",u"pandas",b"pandas",u"pandas_profiling",b"pandas_profiling",u"sklearn",b"sklearn",u"skopt",b"skopt",u"std",b"std",u"tensorflow",b"tensorflow",u"xgboost",b"xgboost"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"imblearn",b"imblearn",u"numpy",b"numpy",u"op",b"op",u"pandas",b"pandas",u"pandas_profiling",b"pandas_profiling",u"sklearn",b"sklearn",u"skopt",b"skopt",u"std",b"std",u"tensorflow",b"tensorflow",u"xgboost",b"xgboost"]) -> None: ...
            def WhichOneof(self, oneof_group: typing_extensions.Literal[u"op",b"op"]) -> typing.Optional[typing_extensions.Literal["std","pandas","numpy","tensorflow","sklearn","pandas_profiling","xgboost","skopt","imblearn"]]: ...

        class Std(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Pandas(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Numpy(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Tensorflow(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Sklearn(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class PandasProfiling(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class XGBoost(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Skopt(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        class Imblearn(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            NAME_FIELD_NUMBER: builtins.int
            name: typing.Text = ...
            def __init__(self,
                *,
                name : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name"]) -> None: ...

        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
        OP_IDENTIFIER_FIELD_NUMBER: builtins.int
        arguments: builtins.bytes = ...
        named_arguments: builtins.bytes = ...
        @property
        def op_identifier(self) -> global___Transform.External.OpIdentifier: ...
        def __init__(self,
            *,
            arguments : builtins.bytes = ...,
            named_arguments : builtins.bytes = ...,
            op_identifier : typing.Optional[global___Transform.External.OpIdentifier] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"op_identifier",b"op_identifier"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"arguments",b"arguments",u"named_arguments",b"named_arguments",u"op_identifier",b"op_identifier"]) -> None: ...

    class Identity(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class Variable(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        POSITION_FIELD_NUMBER: builtins.int
        NAME_FIELD_NUMBER: builtins.int
        position: builtins.int = ...
        name: typing.Text = ...
        def __init__(self,
            *,
            position : builtins.int = ...,
            name : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"name",b"name",u"position",b"position"]) -> None: ...

    class Composed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class NamedArgumentsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text = ...
            value: typing.Text = ...
            def __init__(self,
                *,
                key : typing.Text = ...,
                value : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

        TRANSFORM_FIELD_NUMBER: builtins.int
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
        transform: typing.Text = ...
        """Transform"""

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            """Arguments of the current transform are transforms"""
            pass
        @property
        def named_arguments(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
        def __init__(self,
            *,
            transform : typing.Text = ...,
            arguments : typing.Optional[typing.Iterable[typing.Text]] = ...,
            named_arguments : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"arguments",b"arguments",u"named_arguments",b"named_arguments",u"transform",b"transform"]) -> None: ...

    class Project(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        PROJECTION_FIELD_NUMBER: builtins.int
        @property
        def projection(self) -> sarus_data_spec.protobuf.type_pb2.Type:
            """This should be a 'supertype' the type the data can project into."""
            pass
        def __init__(self,
            *,
            projection : typing.Optional[sarus_data_spec.protobuf.type_pb2.Type] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"projection",b"projection"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"projection",b"projection"]) -> None: ...

    class Filter(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        FILTER_FIELD_NUMBER: builtins.int
        @property
        def filter(self) -> sarus_data_spec.protobuf.type_pb2.Type:
            """This should be a 'subtype' the type the data can be retricted to."""
            pass
        def __init__(self,
            *,
            filter : typing.Optional[sarus_data_spec.protobuf.type_pb2.Type] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"filter",b"filter"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"filter",b"filter"]) -> None: ...

    class Shuffle(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class Join(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        ON_FIELD_NUMBER: builtins.int
        @property
        def on(self) -> sarus_data_spec.protobuf.type_pb2.Type:
            """This should be a common 'supertype' between tables."""
            pass
        def __init__(self,
            *,
            on : typing.Optional[sarus_data_spec.protobuf.type_pb2.Type] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"on",b"on"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"on",b"on"]) -> None: ...

    class Cast(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        TYPE_FIELD_NUMBER: builtins.int
        @property
        def type(self) -> sarus_data_spec.protobuf.type_pb2.Type:
            """Type to cast into."""
            pass
        def __init__(self,
            *,
            type : typing.Optional[sarus_data_spec.protobuf.type_pb2.Type] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"type",b"type"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"type",b"type"]) -> None: ...

    class Sample(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        FRACTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        SEED_FIELD_NUMBER: builtins.int
        fraction: builtins.float = ...
        size: builtins.int = ...
        @property
        def seed(self) -> sarus_data_spec.protobuf.scalar_pb2.Scalar: ...
        def __init__(self,
            *,
            fraction : builtins.float = ...,
            size : builtins.int = ...,
            seed : typing.Optional[sarus_data_spec.protobuf.scalar_pb2.Scalar] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"fraction",b"fraction",u"proportion",b"proportion",u"seed",b"seed",u"size",b"size"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"fraction",b"fraction",u"proportion",b"proportion",u"seed",b"seed",u"size",b"size"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"proportion",b"proportion"]) -> typing.Optional[typing_extensions.Literal["fraction","size"]]: ...

    class SchemaInference(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class CastPolicy(_CastPolicy, metaclass=_CastPolicyEnumTypeWrapper):
            pass
        class _CastPolicy:
            V = typing.NewType('V', builtins.int)
        class _CastPolicyEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_CastPolicy.V], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
            NONE = Transform.SchemaInference.CastPolicy.V(0)
            MOST_LIKELY = Transform.SchemaInference.CastPolicy.V(1)

        NONE = Transform.SchemaInference.CastPolicy.V(0)
        MOST_LIKELY = Transform.SchemaInference.CastPolicy.V(1)

        CAST_POLICY_FIELD_NUMBER: builtins.int
        cast_policy: global___Transform.SchemaInference.CastPolicy.V = ...
        def __init__(self,
            *,
            cast_policy : global___Transform.SchemaInference.CastPolicy.V = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"cast_policy",b"cast_policy"]) -> None: ...

    class GroupBy(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key"]) -> None: ...

    class Synthetic(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class UserSettings(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class AutomaticUserSettings(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        MAX_CATEGORIES_FIELD_NUMBER: builtins.int
        max_categories: builtins.int = ...
        def __init__(self,
            *,
            max_categories : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"max_categories",b"max_categories"]) -> None: ...

    class Protect(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class Transcode(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class InverseTranscode(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class DifferentiatedSample(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        FRACTION_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        SEED_FIELD_NUMBER: builtins.int
        fraction: builtins.float = ...
        size: builtins.int = ...
        @property
        def seed(self) -> sarus_data_spec.protobuf.scalar_pb2.Scalar: ...
        def __init__(self,
            *,
            fraction : builtins.float = ...,
            size : builtins.int = ...,
            seed : typing.Optional[sarus_data_spec.protobuf.scalar_pb2.Scalar] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"fraction",b"fraction",u"proportion",b"proportion",u"seed",b"seed",u"size",b"size"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"fraction",b"fraction",u"proportion",b"proportion",u"seed",b"seed",u"size",b"size"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"proportion",b"proportion"]) -> typing.Optional[typing_extensions.Literal["fraction","size"]]: ...

    class ProtectedPaths(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class PublicPaths(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class GetItem(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        PATH_FIELD_NUMBER: builtins.int
        @property
        def path(self) -> sarus_data_spec.protobuf.path_pb2.Path: ...
        def __init__(self,
            *,
            path : typing.Optional[sarus_data_spec.protobuf.path_pb2.Path] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"path",b"path"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"path",b"path"]) -> None: ...

    class AssignBudget(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class AutomaticBudget(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class AttributesBudget(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class SDBudget(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class DeriveSeed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        RANDOM_INTEGER_FIELD_NUMBER: builtins.int
        random_integer: builtins.int = ...
        def __init__(self,
            *,
            random_integer : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"random_integer",b"random_integer"]) -> None: ...

    class GroupByPE(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class SamplingRatios(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class RelationshipSpec(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    class SelectSql(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class SQLDialect(_SQLDialect, metaclass=_SQLDialectEnumTypeWrapper):
            pass
        class _SQLDialect:
            V = typing.NewType('V', builtins.int)
        class _SQLDialectEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SQLDialect.V], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
            NONE = Transform.SelectSql.SQLDialect.V(0)
            POSTGRES = Transform.SelectSql.SQLDialect.V(1)
            SQL_SERVER = Transform.SelectSql.SQLDialect.V(2)
            MY_SQL = Transform.SelectSql.SQLDialect.V(3)
            SQLLITE = Transform.SelectSql.SQLDialect.V(4)
            ORACLE = Transform.SelectSql.SQLDialect.V(5)
            BIG_QUERY = Transform.SelectSql.SQLDialect.V(6)
            REDSHIFT = Transform.SelectSql.SQLDialect.V(7)
            HIVE = Transform.SelectSql.SQLDialect.V(8)

        NONE = Transform.SelectSql.SQLDialect.V(0)
        POSTGRES = Transform.SelectSql.SQLDialect.V(1)
        SQL_SERVER = Transform.SelectSql.SQLDialect.V(2)
        MY_SQL = Transform.SelectSql.SQLDialect.V(3)
        SQLLITE = Transform.SelectSql.SQLDialect.V(4)
        ORACLE = Transform.SelectSql.SQLDialect.V(5)
        BIG_QUERY = Transform.SelectSql.SQLDialect.V(6)
        REDSHIFT = Transform.SelectSql.SQLDialect.V(7)
        HIVE = Transform.SelectSql.SQLDialect.V(8)

        class AliasedQueries(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            ALIASED_QUERY_FIELD_NUMBER: builtins.int
            @property
            def aliased_query(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Transform.SelectSql.AliasedQuery]: ...
            def __init__(self,
                *,
                aliased_query : typing.Optional[typing.Iterable[global___Transform.SelectSql.AliasedQuery]] = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"aliased_query",b"aliased_query"]) -> None: ...

        class AliasedQuery(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            PATH_FIELD_NUMBER: builtins.int
            QUERY_FIELD_NUMBER: builtins.int
            @property
            def path(self) -> sarus_data_spec.protobuf.path_pb2.Path: ...
            query: typing.Text = ...
            def __init__(self,
                *,
                path : typing.Optional[sarus_data_spec.protobuf.path_pb2.Path] = ...,
                query : typing.Text = ...,
                ) -> None: ...
            def HasField(self, field_name: typing_extensions.Literal[u"path",b"path"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing_extensions.Literal[u"path",b"path",u"query",b"query"]) -> None: ...

        QUERY_FIELD_NUMBER: builtins.int
        ALIASED_QUERIES_FIELD_NUMBER: builtins.int
        SQL_DIALECT_FIELD_NUMBER: builtins.int
        query: typing.Text = ...
        @property
        def aliased_queries(self) -> global___Transform.SelectSql.AliasedQueries: ...
        sql_dialect: global___Transform.SelectSql.SQLDialect.V = ...
        def __init__(self,
            *,
            query : typing.Text = ...,
            aliased_queries : typing.Optional[global___Transform.SelectSql.AliasedQueries] = ...,
            sql_dialect : global___Transform.SelectSql.SQLDialect.V = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"aliased_queries",b"aliased_queries",u"query",b"query",u"select",b"select"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"aliased_queries",b"aliased_queries",u"query",b"query",u"select",b"select",u"sql_dialect",b"sql_dialect"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal[u"select",b"select"]) -> typing.Optional[typing_extensions.Literal["query","aliased_queries"]]: ...

    class Extract(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        SIZE_FIELD_NUMBER: builtins.int
        size: builtins.int = ...
        def __init__(self,
            *,
            size : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"size",b"size"]) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    INVERSIBLE_FIELD_NUMBER: builtins.int
    SCHEMA_PRESERVING_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    """A dataset transform
    e.g. RFC 4122 id used to refer to the transform
    """

    name: typing.Text = ...
    doc: typing.Text = ...
    @property
    def spec(self) -> global___Transform.Spec: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    inversible: builtins.bool = ...
    schema_preserving: builtins.bool = ...
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        name : typing.Text = ...,
        doc : typing.Text = ...,
        spec : typing.Optional[global___Transform.Spec] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        inversible : builtins.bool = ...,
        schema_preserving : builtins.bool = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"spec",b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"doc",b"doc",u"inversible",b"inversible",u"name",b"name",u"properties",b"properties",u"schema_preserving",b"schema_preserving",u"spec",b"spec",u"uuid",b"uuid"]) -> None: ...
global___Transform = Transform



from __future__ import annotations

from typing import Any

try:
    from sklearn import (
        compose,
        impute,
        inspection,
        linear_model,
        metrics,
        model_selection,
        pipeline,
        preprocessing,
    )
except ModuleNotFoundError:
    pass  # error message in typing.py


async def model_fit(model: Any, *args: Any, **kwargs: Any) -> Any:
    fitted_model = model.fit(*args, **kwargs)
    return fitted_model


async def sk_get_n_splits(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.split(*args, **kwargs)


async def sk_inverse_transform(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.inverse_transform(*args, **kwargs)


async def sk_predict(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.predict(*args, **kwargs)


async def sk_predict_log_proba(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.predict_log_proba(*args, **kwargs)


async def sk_predict_proba(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.predict_proba(*args, **kwargs)


async def sk_split(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.split(*args, **kwargs)


async def sk_score(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.score(*args, **kwargs)


async def sk_transform(model: Any, *args: Any, **kwargs: Any) -> Any:
    return model.transform(*args, **kwargs)


async def sk_scale(val: Any, *args: Any, **kwargs: Any) -> Any:
    return preprocessing.scale(val, *args, **kwargs)


async def sk_cross_val_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return model_selection.cross_val_score(*args, **kwargs)


async def sk_train_test_split(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return model_selection.train_test_split(*args, **kwargs)


# metrics


async def sk_accuracy_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.accuracy_score(*args, **kwargs)


async def sk_average_precision_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.average_precision_score(*args, **kwargs)


async def sk_classification_report(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.classification_report(*args, **kwargs)


async def sk_confusion_matrix(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.confusion_matrix(*args, **kwargs)


async def sk_f1_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.f1_score(*args, **kwargs)


async def sk_plot_confusion_matrix(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.plot_confusion_matrix(*args, **kwargs)


async def sk_precision_recall_curve(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.precision_recall_curve(*args, **kwargs)


async def sk_precision_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.precision_score(*args, **kwargs)


async def sk_recall_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.recall_score(*args, **kwargs)


async def sk_roc_auc_score(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.roc_auc_score(*args, **kwargs)


async def sk_auc(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.auc(*args, **kwargs)


async def sk_roc_curve(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.roc_curve(*args, **kwargs)


async def sk_plot_roc_curve(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.plot_roc_curve(*args, **kwargs)


async def sk_plot_precision_recall_curve(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return metrics.plot_precision_recall_curve(*args, **kwargs)


async def sk_label_encoder_fit_transform(
    model: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    X_transformed = model.fit_transform(*args, **kwargs)
    return X_transformed


async def sk_permutation_importance(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return inspection.permutation_importance(*args, **kwargs)


async def sk_pipeline(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return pipeline.Pipeline(*args, **kwargs)


async def sk_function_transformer(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return preprocessing.FunctionTransformer(*args, **kwargs)


async def sk_column_transformer(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return compose.ColumnTransformer(*args, **kwargs)


async def sk_time_series_split(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return model_selection.TimeSeriesSplit(*args, **kwargs)


async def sk_simple_imputer(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return impute.SimpleImputer(*args, **kwargs)


async def sk_linear_regression(
    *args: Any,
    **kwargs: Any,
) -> Any:
    return linear_model.LinearRegression(*args, **kwargs)



from typing import Any, List, Tuple, Union
import typing as t

import numpy as np
import pandas as pd
import pandas._typing as pdt

from sarus_data_spec.dataspec_validator.typing import PEPKind
import sarus_data_spec.typing as st

from ..protection_utils import ExternalOpImplementation
from .pandas_dp import (
    pd_mean_dp,
    pd_median_dp,
    pd_shape_dp,
    pd_std_dp,
    pd_sum_dp,
)

# Defined in pandas version > 1.3.5
IgnoreRaise = t.Literal["ignore", "raise"]


class pd_loc(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        key: Tuple[Union[str, slice, List[str]], ...],
    ) -> pd.DataFrame:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.loc[key]

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        key: Tuple[Union[str, slice, List[str]], ...],
    ) -> PEPKind:
        return PEPKind.PEP


class pd_iloc(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        key: Tuple[Union[str, slice, List[str]], ...],
    ) -> pd.DataFrame:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.iloc[key]

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        key: Tuple[Union[str, slice, List[str]], ...],
    ) -> PEPKind:
        return PEPKind.PEP


class pd_head(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        n: int = 5,
    ) -> Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.head(n=n)

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        n: int = 5,
    ) -> PEPKind:
        return PEPKind.PEP


class pd_astype(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        dtype: pdt.Dtype,
        copy: bool = True,
        errors: IgnoreRaise = "raise",
    ) -> pd.DataFrame:
        return parent_df.astype(dtype=dtype, copy=copy, errors=errors)

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        dtype: pdt.Dtype,
        copy: bool = True,
        errors: IgnoreRaise = "raise",
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_getitem(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame], key: t.Any
    ) -> t.Any:
        return parent_df[key]

    def pep_kind(  # type: ignore[override]
        self, parent_df: t.Union[pd.Series, pd.DataFrame], key: t.Any
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_shape(ExternalOpImplementation):
    _dp_equivalent = pd_shape_dp

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame]
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.shape


class pd_sum(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]
    _dp_equivalent = pd_sum_dp

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        axis: t.Optional[pdt.Axis] = None,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        min_count: int = 0,
    ) -> Any:
        assert type(parent_df) in [pd.DataFrame, pd.Series]
        return parent_df.sum(
            axis=axis,
            level=level,
            skipna=skipna,
            numeric_only=numeric_only,
            min_count=min_count,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        axis: t.Optional[pdt.Axis] = None,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_mean(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]
    _dp_equivalent = pd_mean_dp

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: bool = False,
    ) -> Any:
        assert type(parent_df) in [pd.DataFrame, pd.Series]
        return parent_df.mean(
            axis=axis,
            skipna=skipna,
            level=level,
            numeric_only=numeric_only,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: bool = False,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_std(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]
    _dp_equivalent = pd_std_dp

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        ddof: int = 1,
        numeric_only: bool = False,
    ) -> t.Any:
        assert type(parent_df) in [pd.DataFrame, pd.Series]
        return parent_df.std(
            axis=axis,
            skipna=skipna,
            level=level,
            ddof=ddof,
            numeric_only=numeric_only,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: t.Any,
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        ddof: int = 1,
        numeric_only: bool = False,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_median(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]
    _dp_equivalent = pd_median_dp

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[int] = None,
        numeric_only: bool = False,
    ) -> t.Any:
        assert type(parent_df) in [pd.DataFrame, pd.Series]
        return parent_df.median(
            axis=axis,
            skipna=skipna,
            level=level,
            numeric_only=numeric_only,
        )

    def pep_kind(  # type: ignore[override]
        self,
        dataframe: Any,
        axis: int = 0,
        skipna: bool = True,
        level: t.Optional[int] = None,
        numeric_only: bool = False,
        **kwargs: Any,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_abs(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame]
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.abs()

    def pep_kind(  # type: ignore[override]
        self, parent_df: st.DataSpec
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_drop(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Any,
        labels: pdt.IndexLabel = None,
        axis: pdt.Axis = 0,
        index: pdt.IndexLabel = None,
        columns: pdt.IndexLabel = None,
        level: pdt.Level = None,
        inplace: bool = False,
        errors: IgnoreRaise = "raise",
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.drop(
            axis=axis,
            labels=labels,
            index=index,
            columns=columns,
            level=level,
            inplace=False,  # need to be False or returns None
            errors=errors,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: t.Any,
        labels: pdt.IndexLabel = None,
        axis: pdt.Axis = 0,
        index: pdt.IndexLabel = None,
        columns: pdt.IndexLabel = None,
        level: pdt.Level = None,
        inplace: bool = False,
        errors: IgnoreRaise = "raise",
    ) -> PEPKind:
        if axis in [0, 'columns']:
            return PEPKind.PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_dropna(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Any,
        axis: pdt.Axis = 0,
        how: t.Optional[str] = None,
        thresh: t.Optional[int] = None,
        subset: pdt.IndexLabel = None,
        inplace: bool = False,
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.dropna(
            axis=axis, how=how, thresh=thresh, subset=subset, inplace=inplace
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: t.Any,
        axis: pdt.Axis = 0,
        how: t.Optional[str] = None,
        thresh: t.Optional[int] = None,
        subset: pdt.IndexLabel = None,
        inplace: bool = False,
    ) -> PEPKind:
        if axis in [0, 'columns']:
            return PEPKind.PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_fillna(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        value: t.Optional[
            t.Union[t.Hashable, t.Mapping, pd.Series, pd.DataFrame]
        ] = None,
        method: t.Optional[pdt.FillnaOptions] = None,
        axis: t.Optional[pdt.Axis] = None,
        inplace: bool = False,
        limit: t.Optional[int] = None,
        downcast: t.Optional[dict] = None,
    ) -> Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.fillna(
            value=value,
            method=method,
            axis=axis,
            inplace=False,  # need to be False or returns None
            limit=limit,
            downcast=downcast,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.Dataset,
        value: t.Optional[
            t.Union[t.Hashable, t.Mapping, pd.Series, pd.DataFrame]
        ] = None,
        method: t.Optional[pdt.FillnaOptions] = None,
        axis: t.Optional[pdt.Axis] = None,
        inplace: bool = False,
        limit: t.Optional[int] = None,
        downcast: t.Optional[dict] = None,
    ) -> PEPKind:
        if method is None:
            return PEPKind.TOKEN_PRESERVING
        else:
            return PEPKind.NOT_PEP


class pd_isin(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        values: t.Union[pd.Series, pd.DataFrame, pdt.Sequence, t.Mapping],
    ) -> t.Any:
        return parent_df.isin(values=values)

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.Dataset,
        values: t.Union[pd.Series, pd.DataFrame, pdt.Sequence, t.Mapping],
    ) -> t.Any:
        return PEPKind.TOKEN_PRESERVING


class pd_isnull(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame]
    ) -> Any:
        return parent_df.isnull()

    def pep_kind(  # type: ignore[override]
        self, parent_df: st.Dataset
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_mask(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        cond: t.Union[pd.Series, pd.DataFrame, t.Callable],
        other: t.Union[
            pdt.Scalar, pd.Series, pd.DataFrame, t.Callable
        ] = np.nan,
        inplace: bool = False,
        axis: t.Optional[pdt.Axis] = None,
        level: pdt.Level = None,
        errors: IgnoreRaise = "raise",
        try_cast: bool = False,
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.mask(
            cond=cond,
            other=other,
            inplace=False,  # need to be False or returns None
            axis=axis,
            level=level,
            errors=errors,
            try_cast=try_cast,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.Dataset,
        cond: t.Union[pd.Series, pd.DataFrame, t.Callable],
        other: t.Union[
            pdt.Scalar, pd.Series, pd.DataFrame, t.Callable
        ] = np.nan,
        inplace: bool = False,
        axis: t.Optional[pdt.Axis] = None,
        level: pdt.Level = None,
        errors: IgnoreRaise = "raise",
        try_cast: bool = False,
    ) -> PEPKind:
        if callable(cond) or callable(other):
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_notnull(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame]
    ) -> Any:
        return parent_df.notnull()

    def pep_kind(  # type: ignore[override]
        self, parent_df: st.Dataset
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_rename(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        mapper: t.Optional[pdt.Renamer] = None,
        index: t.Optional[pdt.Renamer] = None,
        columns: t.Optional[pdt.Renamer] = None,
        axis: t.Optional[pdt.Axis] = None,
        copy: t.Optional[bool] = None,
        inplace: bool = False,
        level: pdt.Level = None,
        errors: IgnoreRaise = "ignore",
    ) -> Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.rename(
            mapper=mapper,
            index=index,
            columns=columns,
            axis=axis,
            copy=copy,
            inplace=False,  # need to be False or returns None
            level=level,
            errors=errors,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.Dataset,
        mapper: t.Optional[pdt.Renamer] = None,
        index: t.Optional[pdt.Renamer] = None,
        columns: t.Optional[pdt.Renamer] = None,
        axis: t.Optional[pdt.Axis] = None,
        copy: t.Optional[bool] = None,
        inplace: bool = False,
        level: pdt.Level = None,
        errors: IgnoreRaise = "ignore",
    ) -> PEPKind:
        if callable(mapper) or callable(index) or callable(columns):
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_replace(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        to_replace: t.Union[
            str, t.List, t.Dict, pd.Series, int, float, None
        ] = None,
        value: t.Optional[t.Union[pdt.Scalar, t.Dict, t.List, str]] = None,
        inplace: bool = False,
        limit: t.Optional[int] = None,
        regex: bool = False,
        method: t.Optional[t.Literal["pad", "ffill", "bfill"]] = None,
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.replace(
            to_replace=to_replace,
            value=value,
            inplace=False,  # need to be False or returns None
            limit=limit,
            regex=regex,
            method=method,
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: t.Union[pd.Series, pd.DataFrame],
        to_replace: t.Union[
            str, t.List, t.Dict, pd.Series, int, float, None
        ] = None,
        value: t.Optional[t.Union[pdt.Scalar, t.Dict, t.List, str]] = None,
        inplace: bool = False,
        limit: t.Optional[int] = None,
        regex: bool = False,
        method: t.Optional[t.Literal["pad", "ffill", "bfill"]] = None,
    ) -> PEPKind:
        if value is None:
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_round(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        decimals: t.Union[int, t.Dict[pdt.IndexLabel, int], pd.Series] = 0,
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.round(decimals=decimals)

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.Dataset,
        decimals: t.Union[int, t.Dict[pdt.IndexLabel, int], pd.Series] = 0,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_select_dtypes(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        include: t.Optional[t.Union[pdt.Scalar, t.List]] = None,
        exclude: t.Optional[t.Union[pdt.Scalar, t.List]] = None,
    ) -> t.Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.select_dtypes(include=include, exclude=exclude)

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        include: t.Optional[t.Union[pdt.Scalar, t.List]] = None,
        exclude: t.Optional[t.Union[pdt.Scalar, t.List]] = None,
    ) -> PEPKind:
        return PEPKind.PEP


class pd_add(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        other: t.Union[
            pdt.Scalar, t.Sequence, pd.Series, t.Dict, pd.DataFrame
        ],
        axis: pdt.Axis = 'columns',
        level: t.Optional[pdt.Level] = None,
        fill_value: t.Optional[float] = None,
    ) -> Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.add(
            other=other, axis=axis, level=level, fill_value=fill_value
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        other: t.Union[
            pdt.Scalar, t.Sequence, pd.Series, t.Dict, pd.DataFrame
        ],
        axis: pdt.Axis = 'columns',
        level: t.Optional[pdt.Level] = None,
        fill_value: t.Optional[float] = None,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_sub(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_df"}]

    @staticmethod
    async def call(  # type: ignore[override]
        parent_df: t.Union[pd.Series, pd.DataFrame],
        other: t.Union[
            pdt.Scalar, t.Sequence, pd.Series, t.Dict, pd.DataFrame
        ],
        axis: pdt.Axis = 'columns',
        level: t.Optional[pdt.Level] = None,
        fill_value: t.Optional[float] = None,
    ) -> Any:
        assert type(parent_df) in [pd.Series, pd.DataFrame]
        return parent_df.sub(
            other=other, axis=axis, level=level, fill_value=fill_value
        )

    def pep_kind(  # type: ignore[override]
        self,
        parent_df: st.DataSpec,
        other: t.Union[
            pdt.Scalar, t.Sequence, pd.Series, t.Dict, pd.DataFrame
        ],
        axis: pdt.Axis = 'columns',
        level: t.Optional[pdt.Level] = None,
        fill_value: t.Optional[float] = None,
    ) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


async def pd_sum_groupby(groupby: t.Any, *args: t.Any, **kwargs: t.Any) -> Any:
    """SUM operation specific for groupby objects."""
    assert isinstance(groupby, pd.core.groupby.GroupBy)
    return groupby.sum(*args, **kwargs)


async def pd_std_groupby(groupby: t.Any, *args: t.Any, **kwargs: t.Any) -> Any:
    """STD operation specific for groupby objects."""
    assert isinstance(groupby, pd.core.groupby.GroupBy)
    return groupby.std(*args, **kwargs)


async def pd_mean_groupby(
    groupby: t.Any, *args: t.Any, **kwargs: t.Any
) -> Any:
    """MEAN operation specific for groupby objects."""
    assert isinstance(groupby, pd.core.groupby.GroupBy)
    return groupby.mean(*args, **kwargs)


async def pd_median_groupby(
    groupby: t.Any, *args: t.Any, **kwargs: t.Any
) -> Any:
    """MEDIAN operation specific for groupby objects."""
    assert isinstance(groupby, pd.core.groupby.GroupBy)
    return groupby.median(*args, **kwargs)


async def pd_ndim(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.ndim


async def pd_dataframe(*args: Any, **kwargs: Any) -> Any:
    return pd.DataFrame(*args, **kwargs)


async def pd_series(*args: Any, **kwargs: Any) -> Any:
    return pd.Series(*args, **kwargs)


async def pd_query(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.DataFrame]
    return parent_val.query(*args, **kwargs)


async def pd_groups(parent_val: Any) -> Any:
    return parent_val.groups


async def pd_name(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.name


async def pd_size(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.size


async def pd_axes(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.axes


async def pd_columns(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.columns


async def pd_index(parent_val: Any) -> pd.DataFrame:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.index


async def pd_dtype(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.dtype


async def pd_dtypes(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.dtypes


async def pd_values(parent_val: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.values


async def pd_reset_index(
    parent_val: Any, *args: Any, **kwargs: Any
) -> pd.DataFrame:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.reset_index(*args, **kwargs)


async def pd_set_loc(
    parent_val: Any,
    key: Tuple[Union[str, slice, List[str]], ...],
    newvalue: Any,
) -> pd.DataFrame:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    parent_val.loc[key] = newvalue
    return parent_val


async def pd_set_iloc(
    parent_val: Any,
    key: Tuple[Union[str, slice, List[str]], ...],
    newvalue: Any,
) -> pd.DataFrame:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    parent_val.iloc[key] = newvalue
    return parent_val


async def pd_eq(val_1: Any, val_2: Any) -> pd.DataFrame:
    return val_1 == val_2


async def pd_min(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.min(*args, **kwargs)


async def pd_max(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.max(*args, **kwargs)


async def pd_shift(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.shift(*args, **kwargs)


async def pd_any(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.any(*args, **kwargs)


async def pd_describe(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.describe(*args, **kwargs)


async def pd_quantile(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.quantile(*args, **kwargs)


async def pd_reindex(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.reindex(*args, **kwargs)


async def pd_count(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.count(*args, **kwargs)


async def pd_transpose(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.transpose(*args, **kwargs)


async def pd_unique(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.unique(*args, **kwargs)


async def pd_value_counts(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.value_counts(*args, **kwargs)


async def pd_to_dict(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.to_dict(*args, **kwargs)


async def pd_apply(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.apply(*args, **kwargs)


async def pd_mad(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.mad(*args, **kwargs)


async def pd_skew(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.skew(*args, **kwargs)


async def pd_kurtosis(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.kurtosis(*args, **kwargs)


async def pd_agg(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.agg(*args, **kwargs)


async def pd_droplevel(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.droplevel(*args, **kwargs)


async def pd_sort_values(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.sort_values(*args, **kwargs)


async def pd_drop_duplicates(
    parent_val: Any, *args: Any, **kwargs: Any
) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.drop_duplicates(*args, **kwargs)


async def pd_corr(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return parent_val.corr(*args, **kwargs)


async def pd_get_dummies(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    assert type(parent_val) in [pd.Series, pd.DataFrame]
    return pd.get_dummies(parent_val, *args, **kwargs)


async def pd_join(val1: Any, *args: Any, **kwargs: Any) -> Any:
    return val1.join(*args, **kwargs)


async def pd_groupby(val1: Any, *args: Any, **kwargs: Any) -> Any:
    return val1.groupby(*args, **kwargs)


async def pd_merge(val1: Any, *args: Any, **kwargs: Any) -> Any:
    return val1.merge(*args, **kwargs)


async def pd_merge_fn(*args: Any, **kwargs: Any) -> Any:
    return pd.merge(*args, **kwargs)


async def pd_concat(*args: Any, **kwargs: Any) -> Any:
    return pd.concat(*args, **kwargs)


async def pd_append(val1: Any, *args: Any, **kwargs: Any) -> Any:
    return val1.append(*args, **kwargs)


async def pd_nunique(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.nunique(*args, **kwargs)


async def pd_rolling(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.rolling(*args, **kwargs)


async def pd_union(parent_val: Any, *args: Any, **kwargs: Any) -> Any:
    return parent_val.union(*args, **kwargs)


async def pd_to_datetime(*args: Any, **kwargs: Any) -> Any:
    return pd.to_datetime(*args, **kwargs)



import logging
import typing as t

import numpy as np
import pandas as pd
import pandas._typing as pdt

from sarus_data_spec.constants import MAX_MAX_MULT
from sarus_data_spec.protobuf.utilities import unwrap
import sarus_data_spec.typing as st

from ..protection_utils import (
    ExternalOpImplementation,
    arguments_values,
    arguments_values_and_protected_entity,
)

logger = logging.getLogger(__name__)

try:
    from sarus_statistics.ops.bounds.op import BoundOp
    from sarus_statistics.ops.histograms.op import CountOp
    from sarus_statistics.ops.max_multiplicity.op import MaxMultiplicityOp
    from sarus_statistics.ops.mean.op import MeanOp
    from sarus_statistics.ops.median.op import MedianOp
    from sarus_statistics.ops.std.op import StdOp
    from sarus_statistics.ops.sum.op import SumOp
    from sarus_statistics.protobuf.size_pb2 import MaxMultiplicity
except ModuleNotFoundError:
    pass  # warning raised in typing

try:
    from sarus_differential_privacy.query import ComposedPrivateQuery
except ModuleNotFoundError:
    pass  # warning raised in typing

try:
    from sarus_query_builder.builders.bounds_builder import (
        simple_bounds_builder,
    )
    from sarus_query_builder.builders.composed_builder import (
        simple_composed_builder,
    )
    from sarus_query_builder.builders.max_multiplicity_builder import (
        simple_max_multiplicity_builder,
    )
    from sarus_query_builder.builders.mean_builder import mean_builder
    from sarus_query_builder.builders.median_builder import median_builder
    from sarus_query_builder.builders.standard_mechanisms_builder import (
        laplace_builder,
    )
    from sarus_query_builder.builders.std_builder import std_builder
    from sarus_query_builder.builders.sum_builder import sum_builder
    from sarus_query_builder.core.core import OptimizableQueryBuilder
    from sarus_query_builder.protobuf.query_pb2 import GenericTask, Query
except ModuleNotFoundError:
    pass  # warning raised in typing


DEFAULT_MAX_MAX_MULT = 1
NUMERIC_TYPES = ('integer', 'float', 'boolean')


class pd_shape_dp(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_ds"}]
    transform_id: str = "pandas.PD_SHAPE_DP"

    def is_dp(self, *args: t.Any, **kwargs: t.Any) -> bool:
        return True

    @staticmethod
    async def private_queries_and_task(  # type: ignore[override]
        parent_ds: st.Dataset, budget: t.Any, **kwargs: t.Any
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        """Return the PrivateQueries summarizing DP characteristics."""
        if len(budget) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )

        epsilon = budget[0].epsilon
        delta = budget[0].delta
        if epsilon == 0.0:
            raise ValueError("`epsilon` should be greater than 0.")

        builder = simple_composed_builder(
            parent_ds,
            [
                simple_max_multiplicity_builder(
                    parent_ds,
                    Query(
                        max_multiplicity=Query.MaxMultiplicity(
                            max_max_multiplicity=max_max_mult
                        )
                    ),
                ),
                laplace_builder(
                    parent_ds,
                    Query(laplace_mechanism=Query.LaplaceMechanism()),
                ),
            ],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    @staticmethod
    async def call(  # type: ignore[override]
        parent_ds: st.Dataset,
        budget: st.Scalar,
        seed: st.Scalar,
    ) -> t.Any:
        """Implementation of DP shape.

        A DP implementation receives additional arguments compared to a
        standard external implementation:
            - `budget`: a list of sp.Scalar.PrivacyParams.Point
                object containing each an epsilon and a delta values
            - `seed`: an integer used to parametrize random number generators
            - `pe`: theprotected entity used by `sarus_statistics` primitives
        """
        # Evaluate arguments
        (dataframe, budget_value, seed_value), _ = await arguments_values(
            parent_ds, budget, seed
        )

        # Get QB task parametrization
        _, tasks = await pd_shape_dp.private_queries_and_task(
            parent_ds, budget_value
        )

        epsilon = budget_value[0].epsilon
        n_rows, n_cols = dataframe.shape

        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, shape_task = t.cast(MaxMultiplicity, tasks[0]), t.cast(
            GenericTask, tasks[1]
        )
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        n_rows = CountOp(
            parent_ds,
            noise=shape_task.parameters['noise'],
        ).value(max_mul, random_generator)

        dp_shape = (n_rows, n_cols)

        # Compute private query
        return dp_shape


class pd_sum_dp(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_ds"}]
    transform_id: str = "pandas.PD_SUM_DP"

    def is_dp(  # type: ignore[override]
        self,
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = None,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        min_count: int = 0,
        budget: t.Any = None,
        seed: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> bool:
        return (axis == 0) and (numeric_only is True)

    @staticmethod
    async def private_queries_and_task(  # type: ignore[override]
        parent_ds: st.Dataset,
        budget: t.Any,
        **kwargs: t.Any,
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        if len(budget) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )

        epsilon = budget[0].epsilon
        delta = budget[0].delta
        if epsilon == 0.0:
            raise ValueError("`epsilon` should be greater than 0.")

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )

        column_bounds_sum_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    sum_builder(parent_ds, Query(sum=Query.Sum())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_sum_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    @staticmethod
    async def call(  # type: ignore[override]
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = None,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        min_count: int = 0,
        budget: t.Optional[st.Scalar] = None,
        seed: t.Optional[st.Scalar] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        # Evaluate arguments
        (dataframe, budget_value, seed_value), _ = await arguments_values(
            parent_ds, budget, seed
        )

        assert type(dataframe) in [pd.DataFrame, pd.Series]
        assert numeric_only
        assert axis == 0
        assert level is None
        assert min_count == 0
        assert seed_value
        assert budget_value

        _, tasks = await pd_sum_dp.private_queries_and_task(
            parent_ds=parent_ds,
            budget=budget_value,
        )

        epsilon = budget_value[0].epsilon
        # Compute DP value
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, sum_tasks = tasks[0], tasks[1:]
        random_generator = np.random.default_rng(abs(seed_value))
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        sum_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            subtasks = [unwrap(task) for task in sum_tasks[index].subtasks]
            bounds_parameters, sum_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    sum_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            # TODO parent_ds doesn't have a size
            # if parent_ds.size().children()[column_name].size() < min_count:
            #     sum_dp_dict[column_name] = np.nan
            #     continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            sum_op = SumOp(
                parent_ds,
                sum_parameters.parameters['noise'],
            )
            sum_dp_dict[column_name] = sum_op.value(
                column_name, max_mul, bounds, random_generator
            )

        sum_dp = pd.Series(sum_dp_dict)
        return sum_dp


class pd_mean_dp(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_ds"}]
    transform_id: str = "pandas.PD_MEAN_DP"

    def is_dp(  # type: ignore[override]
        self,
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        budget: t.Any = None,
        seed: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> bool:
        return axis == 0 and numeric_only is True

    @staticmethod
    async def private_queries_and_task(  # type: ignore[override]
        parent_ds: st.Dataset,
        budget: t.Any,
        **kwargs: t.Any,
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        if len(budget) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )

        epsilon = budget[0].epsilon
        delta = budget[0].delta
        if epsilon == 0.0:
            raise ValueError("`epsilon` should be greater than 0.")

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_mean_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    mean_builder(parent_ds, Query(mean=Query.Mean())),
                ],
            )
        )
        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_mean_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    @staticmethod
    async def call(  # type: ignore[override]
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        budget: t.Optional[st.Scalar] = None,
        seed: t.Optional[st.Scalar] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        # Evaluate arguments
        (
            (dataframe, budget_value, seed_value),
            _,
            pe,
        ) = await arguments_values_and_protected_entity(
            parent_ds, budget, seed
        )

        assert type(dataframe) in [pd.DataFrame, pd.Series]
        assert pe.shape[0] == dataframe.shape[0]
        assert numeric_only
        assert axis == 0
        assert level is None
        assert seed_value
        assert budget_value

        _, tasks = await pd_mean_dp.private_queries_and_task(
            parent_ds=parent_ds,
            budget=budget_value,
        )
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, mean_tasks = tasks[0], tasks[1:]

        # Compute DP value
        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        mean_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in mean_tasks[index].subtasks]
            bounds_parameters, mean_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    mean_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            mean_op = MeanOp(
                parent_ds,
                mean_parameters.parameters['noise'],
            )
            mean_dp_dict[column_name] = mean_op.value(
                column_name, max_mul, bounds, random_generator
            )

        mean_dp = pd.Series(mean_dp_dict)
        return mean_dp


class pd_median_dp(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_ds"}]
    transform_id: str = "pandas.PD_MEDIAN_DP"

    def is_dp(  # type: ignore[override]
        self,
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        budget: t.Any = None,
        seed: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> bool:
        return axis == 0 and numeric_only is True

    @staticmethod
    async def private_queries_and_task(  # type: ignore[override]
        parent_ds: st.Dataset,
        budget: t.Any,
        **kwargs: t.Any,
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        if len(budget) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )

        epsilon = budget[0].epsilon
        delta = budget[0].delta
        if epsilon == 0.0:
            raise ValueError("`epsilon` should be greater than 0.")

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_median_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    median_builder(parent_ds, Query(median=Query.Median())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_median_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    @staticmethod
    async def call(  # type: ignore[override]
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        budget: t.Optional[st.Scalar] = None,
        seed: t.Optional[st.Scalar] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        # Evaluate arguments
        (
            (dataframe, budget_value, seed_value),
            _,
            pe,
        ) = await arguments_values_and_protected_entity(
            parent_ds, budget, seed
        )

        assert type(dataframe) in [pd.DataFrame, pd.Series]
        assert pe.shape[0] == dataframe.shape[0]
        assert numeric_only
        assert axis == 0
        assert level is None
        assert seed_value
        assert budget_value

        _, tasks = await pd_median_dp.private_queries_and_task(
            parent_ds=parent_ds,
            budget=budget_value,
        )
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, median_tasks = tasks[0], tasks[1:]

        random_generator = np.random.default_rng(abs(seed_value))
        epsilon = budget_value[0].epsilon

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        median_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in median_tasks[index].subtasks]
            bounds_parameters, median_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    median_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            median_op = MedianOp(
                parent_ds,
                median_parameters.parameters['noise'],
            )
            median_dp_dict[column_name] = median_op.value(
                column_name, max_mul, bounds, random_generator
            )

        median_dp = pd.Series(median_dp_dict)
        return median_dp


class pd_std_dp(ExternalOpImplementation):
    allowed_pep_args: t.List[t.Set[str]] = [{"parent_ds"}]
    transform_id: str = "pandas.PD_STD_DP"

    def is_dp(  # type: ignore[override]
        self,
        parent_ds: t.Any,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        ddof: int = 1,
        budget: t.Any = None,
        seed: t.Optional[int] = None,
        **kwargs: t.Any,
    ) -> bool:
        return axis == 0 and numeric_only is True

    @staticmethod
    async def private_queries_and_task(  # type: ignore[override]
        parent_ds: st.Dataset,
        budget: t.Any,
        **kwargs: t.Any,
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        if len(budget) != 1:
            raise NotImplementedError(
                "The PrivacyParams contains more than 1 point in the privacy "
                "profile."
            )

        epsilon = budget[0].epsilon
        delta = budget[0].delta
        if epsilon == 0.0:
            raise ValueError("`epsilon` should be greater than 0.")

        parent_schema = await parent_ds.manager().async_schema(parent_ds)
        max_max_mult = int(
            parent_schema.properties().get(MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT)
        )
        n_cols = len(parent_schema.data_type().children())

        max_mult_builder: OptimizableQueryBuilder = (
            simple_max_multiplicity_builder(
                parent_ds,
                Query(
                    max_multiplicity=Query.MaxMultiplicity(
                        max_max_multiplicity=max_max_mult
                    )
                ),
            )
        )
        column_bounds_std_builder: OptimizableQueryBuilder = (
            simple_composed_builder(
                parent_ds,
                [
                    simple_bounds_builder(
                        parent_ds, Query(bounds=Query.Bounds())
                    ),
                    std_builder(parent_ds, Query(median=Query.Median())),
                ],
            )
        )

        builder = simple_composed_builder(
            parent_ds,
            [max_mult_builder] + n_cols * [column_bounds_std_builder],
        )
        tasks = builder.build_query(
            builder.target([(epsilon, delta)], (0, epsilon))
        )
        query = builder.private_query(tasks)
        composed_query = t.cast(ComposedPrivateQuery, query)
        return list(composed_query.all_subqueries()), tasks

    @staticmethod
    async def call(  # type: ignore[override]
        parent_ds: st.Dataset,
        axis: t.Optional[pdt.Axis] = 0,
        skipna: bool = True,
        level: t.Optional[pdt.Level] = None,
        numeric_only: t.Optional[bool] = None,
        ddof: int = 1,
        budget: t.Optional[st.Scalar] = None,
        seed: t.Optional[st.Scalar] = None,
        **kwargs: t.Any,
    ) -> t.Any:
        # Evaluate arguments
        (
            (dataframe, budget_value, seed_value),
            _,
            pe,
        ) = await arguments_values_and_protected_entity(
            parent_ds, budget, seed
        )

        assert type(dataframe) in [pd.DataFrame, pd.Series]
        assert pe.shape[0] == dataframe.shape[0]
        assert numeric_only
        assert axis == 0
        assert level is None
        assert seed_value
        assert budget_value

        _, tasks = await pd_std_dp.private_queries_and_task(
            parent_ds=parent_ds,
            budget=budget_value,
        )
        tasks = [unwrap(subtask) for subtask in tasks.subtasks]
        max_mult_task, std_tasks = tasks[0], tasks[1:]

        epsilon = budget_value[0].epsilon
        random_generator = np.random.default_rng(abs(seed_value))

        # Compute DP value
        max_mul = MaxMultiplicityOp(
            parent_ds,
            epsilon,  # parameter for quantiles
            max_mult_task.noise_user_count,
            max_mult_task.noise_multiplicity,
            max_mult_task.max_max_multiplicity,
        ).value(random_generator)

        std_dp_dict = {}
        for index, column_name in enumerate(dataframe.columns):
            df = dataframe[[column_name]].join(pe)
            subtasks = [unwrap(task) for task in std_tasks[index].subtasks]
            bounds_parameters, std_parameters = t.cast(
                GenericTask, subtasks[0]
            ), t.cast(GenericTask, subtasks[1])
            column_type = (
                parent_ds.schema()
                .data_type()
                .children()[column_name]
                .protobuf()
            )
            if column_type.WhichOneof('type') == 'optional':
                if not skipna:
                    std_dp_dict[column_name] = np.nan
                    continue
                column_type = column_type.optional.type
                df = df.dropna()
            if column_type.WhichOneof('type') not in NUMERIC_TYPES:
                continue

            bounds = BoundOp(
                parent_ds,
                bounds_parameters.parameters['noise'],
            ).value(column_name, max_mul, random_generator)
            std_op = StdOp(
                parent_ds,
                std_parameters.parameters['noise_mean'],
                std_parameters.parameters['noise_square'],
                std_parameters.parameters['noise_count'],
            )
            std_dp_dict[column_name] = std_op.value(
                column_name, max_mul, bounds, random_generator
            )

        std_dp = pd.Series(std_dp_dict)
        return std_dp



import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import (
    DATASET_SLUGNAME,
    OPTIONAL_VALUE,
    PRIMARY_KEYS,
)
from sarus_data_spec.manager.ops.asyncio.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.manager.ops.asyncio.processor.standard.visitor_selector import (  # noqa: E501
    filter_primary_keys,
    select_columns,
    update_fks,
)
from sarus_data_spec.marginals import marginals as marg_builder
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.statistics as sds
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


class ProjectStaticChecker(StandardDatasetStaticChecker):
    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """This transform doesn't change the PEP alignment."""
        return self.parent().pep_token()

    async def schema(self) -> st.Schema:
        new_type = sdt.Type(
            self.dataset.transform().protobuf().spec.project.projection
        )
        parent_schema = await self.parent_schema()
        new_type = update_fks(
            curr_type=new_type, original_type=new_type  # type:ignore
        )
        old_properties = parent_schema.properties()

        if PRIMARY_KEYS in old_properties.keys():
            new_pks = filter_primary_keys(
                old_properties[PRIMARY_KEYS],
                new_type,
            )
            old_properties[PRIMARY_KEYS] = new_pks  # type:ignore

        return schema(
            self.dataset,
            schema_type=new_type,
            protected_paths=parent_schema.protected_path().protobuf(),
            properties=old_properties,
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Project(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a user_settings transform
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        schema = await self.dataset.manager().async_schema(
            dataset=self.dataset
        )
        parent_schema = await self.parent_schema()

        async def async_generator(
            parent_iter: t.AsyncIterator[pa.RecordBatch],
        ) -> t.AsyncIterator[pa.RecordBatch]:
            async for batch in parent_iter:
                updated_array = select_columns(
                    schema.type(),
                    convert_record_batch(
                        record_batch=batch, _type=parent_schema.type()
                    ),
                )
                yield pa.RecordBatch.from_struct_array(updated_array)

        return async_generator(
            parent_iter=await self.parent_to_arrow(batch_size=batch_size)
        )

    async def size(self) -> st.Size:
        schema = await self.dataset.manager().async_schema(self.dataset)
        sizes = await self.parent_size()
        new_stats = update_statistics(
            stats=sizes.statistics(), new_type=schema.data_type()
        )
        return size_builder(dataset=self.dataset, statistics=new_stats)

    async def bounds(self) -> st.Bounds:
        schema = await self.dataset.manager().async_schema(self.dataset)
        bounds = await self.parent_bounds()
        new_stats = update_statistics(
            stats=bounds.statistics(), new_type=schema.data_type()
        )
        return bounds_builder(dataset=self.dataset, statistics=new_stats)

    async def marginals(self) -> st.Marginals:
        schema = await self.dataset.manager().async_schema(self.dataset)
        marginals = await self.parent_marginals()
        new_stats = update_statistics(
            stats=marginals.statistics(), new_type=schema.data_type()
        )
        return marg_builder(dataset=self.dataset, statistics=new_stats)


def update_statistics(
    stats: st.Statistics, new_type: st.Type
) -> st.Statistics:
    """Visitor to update recursively the stats object via the new_type.
    Sub_statistics whose corresponding type is absent in new_type are removed.
    """

    class Updater(st.StatisticsVisitor):
        result = stats

        def Struct(
            self,
            fields: t.Mapping[str, st.Statistics],
            size: int,
            multiplicity: float,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            # project does affect structs
            children_type = new_type.children()
            children_stats = self.result.children()
            new_struct = sds.Struct(
                fields={
                    fieldname: update_statistics(
                        children_stats[fieldname], fieldtype
                    )
                    for fieldname, fieldtype in children_type.items()
                },
                size=size,
                multiplicity=multiplicity,
                name=name,
                properties=self.result.properties(),
            )
            self.result = new_struct

        def Union(
            self,
            fields: t.Mapping[str, st.Statistics],
            size: int,
            multiplicity: float,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            children_type = new_type.children()
            new_fields = {
                fieldname: update_statistics(
                    fieldstats, children_type[fieldname]
                )
                for fieldname, fieldstats in self.result.children().items()
            }
            self.result = sds.Union(
                fields=new_fields,
                size=size,
                multiplicity=multiplicity,
                name=name if name is not None else 'Union',
                properties=self.result.properties(),
            )

        def Optional(
            self, statistics: st.Statistics, size: int, multiplicity: float
        ) -> None:
            self.result = sds.Optional(
                statistics=update_statistics(
                    self.result.children()[OPTIONAL_VALUE],
                    new_type.children()[OPTIONAL_VALUE],
                ),
                size=size,
                multiplicity=multiplicity,
                properties=self.result.properties(),
            )

        def Null(self, size: int, multiplicity: float) -> None:
            pass

        def Unit(self, size: int, multiplicity: float) -> None:
            pass

        def Boolean(
            self,
            size: int,
            multiplicity: float,
            probabilities: t.Optional[t.List[float]] = None,
            names: t.Optional[t.List[bool]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Id(self, size: int, multiplicity: float) -> None:
            pass

        def Integer(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Enum(
            self,
            size: int,
            multiplicity: float,
            probabilities: t.Optional[t.List[float]] = None,
            names: t.Optional[t.List[str]] = None,
            values: t.Optional[t.List[float]] = None,
            name: str = 'Enum',
        ) -> None:
            pass

        def Float(
            self,
            size: int,
            multiplicity: float,
            min_value: float,
            max_value: float,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[float]] = None,
        ) -> None:
            pass

        def Text(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            example: str = '',
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Bytes(self, size: int, multiplicity: float) -> None:
            pass

        def List(
            self,
            statistics: st.Statistics,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            name: str = 'List',
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Array(
            self,
            statistics: st.Statistics,
            size: int,
            multiplicity: float,
            min_values: t.Optional[t.List[float]] = None,
            max_values: t.Optional[t.List[float]] = None,
            name: str = 'Array',
            probabilities: t.Optional[t.List[t.List[float]]] = None,
            values: t.Optional[t.List[t.List[float]]] = None,
        ) -> None:
            raise NotImplementedError

        def Datetime(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Date(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Time(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Duration(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            pass

        def Constrained(
            self, statistics: st.Statistics, size: int, multiplicity: float
        ) -> None:
            raise NotImplementedError

    visitor = Updater()
    stats.accept(visitor)
    return visitor.result



import logging
import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.type import type_from_arrow
from sarus_data_spec.manager.ops.asyncio.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.schema import schema
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class SelectSQLStaticChecker(StandardDatasetStaticChecker):
    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        return None

    async def schema(self) -> st.Schema:
        """With sqlalchemy is not possible to get type detection on text
        statement queries.
        https://github.com/sqlalchemy/sqlalchemy/issues/5677
        however, depending on the DBAPI used we can retrive its type descrption
        from the cursor. For those DBAPI not having this feature we are forced
        to execute the query and retrieve the schema from the results.

        Since the schema is public we may want to execute the query on the
        sythetic dataset.
        """
        select_sql_proto = self.dataset.transform().protobuf().spec.select_sql
        if select_sql_proto.query != '':
            parent_ds = t.cast(st.Dataset, self.parent(kind='dataset'))
            parent_schema = await self.parent_schema()
            logger.debug(
                f"Query {select_sql_proto.query} sent to the "
                f"parent dataset {parent_ds.uuid()} for the schema computation"
            )
            res = await parent_ds.manager().async_sql(
                dataset=parent_ds,
                query=select_sql_proto.query,
                dialect=t.cast(st.SQLDialect, select_sql_proto.sql_dialect),
            )
            table = pa.Table.from_batches(
                batches=[batch async for batch in res]
            )
            fields = {
                col: type_from_arrow(
                    arrow_type=table.field(col).type,
                    nullable=table.field(col).nullable,
                )
                for col in table.schema.names
            }
            schema_type = sdt.Struct(fields=fields)
            return schema(
                self.dataset,
                schema_type=schema_type,
                name=parent_schema.name(),
            )

        else:
            raise NotImplementedError(
                "SelectSQL is not implemented for a dict of queries"
            )


class SelectSQL(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a select_sql transform
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        select_sql_proto = self.dataset.transform().protobuf().spec.select_sql
        if select_sql_proto.query != '':
            parent_ds = t.cast(st.Dataset, self.parent(kind='dataset'))
            logger.debug(
                f"Query {select_sql_proto.query} sent to the "
                f"parent dataset {parent_ds.uuid()} for to arrow"
            )
            return await parent_ds.manager().async_sql(
                dataset=parent_ds,
                query=select_sql_proto.query,
                dialect=t.cast(st.SQLDialect, select_sql_proto.sql_dialect),
                batch_size=batch_size,
            )
        raise NotImplementedError(
            "SelectSQL is not implemented for a dict of queries"
        )

    # TODO
    # def sql_implementation(self) -> t.Optional[t.Dict[st.Path, str]]:
    #     query = self.dataset.transform().protobuf().spec.select_sql.query
    #     aliased_queries: t.Optional[t.Dict[st.Path, str]]
    #     if query == '':
    #         aliased_queries_proto = (
    #             self.dataset.transform()
    #             .protobuf()
    #             .spec.select_sql.aliased_queries
    #         )
    #         aliased_queries = {
    #             Path(alias_q.path): alias_q.query
    #             for alias_q in aliased_queries_proto.aliased_query
    #         }
    #     else:
    #         table_paths = self.dataset.schema().tables()
    #         assert len(table_paths) == 1
    #         aliased_queries = {table_paths[0]: query}
    #     return aliased_queries
